package rs.dmi.master.game;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

import rs.dmi.master.component.edge.Edge;
import rs.dmi.master.component.edge.EdgeCollection;
import rs.dmi.master.component.vertex.Vertex;

/**
 * Complete graph.
 * 
 * @author niktrk
 * @date 20.10.2012.
 * 
 */
public class CompleteGameGraph {

	private final int numberOfVertices;
	private final List<Vertex> vertexList;
	private final EdgeCollection makersEdges;
	private final EdgeCollection breakersEdges;
	private final EdgeCollection freeEdges;
	private final EdgeCollection usedEdges;

	private final GamePlay gamePlay;

	public CompleteGameGraph(int numberOfVertices, GamePlay gamePlay, Edge... excludeEdges) {
		this.numberOfVertices = numberOfVertices;
		this.vertexList = new ArrayList<Vertex>();
		for (int i = 0; i < numberOfVertices; i++) {
			vertexList.add(new Vertex(i));
		}
		this.makersEdges = new EdgeCollection();
		this.breakersEdges = new EdgeCollection();
		this.usedEdges = new EdgeCollection();
		this.freeEdges = new EdgeCollection();
		for (int i = 1; i < numberOfVertices; i++) {
			for (int j = 0; j < i; j++) {
				freeEdges.add(Edge.EDGES[i][j]);
			}
		}
		for (Edge edge : excludeEdges) {
			freeEdges.remove(edge);
		}
		this.gamePlay = gamePlay;
	}

	public void useEdge(Edge edge, Player player) {
		freeEdges.remove(edge);
		usedEdges.add(edge);
		if (Player.MAKER.equals(player)) {
			makersEdges.add(edge);
			vertexList.get(edge.getFirstVertex()).incMakerDeg();
			vertexList.get(edge.getSecondVertex()).incMakerDeg();
		} else {
			breakersEdges.add(edge);
			vertexList.get(edge.getFirstVertex()).incBreakerDeg();
			vertexList.get(edge.getSecondVertex()).incBreakerDeg();
		}
	}

	public void freeEdge(Edge edge, Player player) {
		freeEdges.add(edge);
		usedEdges.remove(edge);
		if (Player.MAKER.equals(player)) {
			makersEdges.remove(edge);
			vertexList.get(edge.getFirstVertex()).decMakerDeg();
			vertexList.get(edge.getSecondVertex()).decMakerDeg();
		} else {
			breakersEdges.remove(edge);
			vertexList.get(edge.getFirstVertex()).decBreakerDeg();
			vertexList.get(edge.getSecondVertex()).decBreakerDeg();
		}
	}

	public int getNumberOfVertices() {
		return numberOfVertices;
	}

	public EdgeCollection getMakersEdges() {
		return makersEdges;
	}

	public EdgeCollection getBreakersEdges() {
		return breakersEdges;
	}

	public EdgeCollection getFreeEdges() {
		return freeEdges;
	}

	public EdgeCollection getUsedEdges() {
		return usedEdges;
	}

	public Player getStartPlayer() {
		return gamePlay.getFirstPlayer();
	}

	public Player getNextPlayer() {
		return gamePlay.getNextPlayer(usedEdges.getSize());
	}

	public List<Vertex> getVertexList() {
		return vertexList;
	}

	public Iterable<Edge> getFreeEdges(Comparator<Edge> comparator) {
		if (comparator == null) {
			return freeEdges;
		} else {
			Edge[] sorted = new Edge[freeEdges.getSize()];
			int i = 0;
			for (Edge freeEdge : freeEdges) {
				sorted[i] = freeEdge;
				i++;
			}
			Arrays.sort(sorted, comparator);
			return Arrays.asList(sorted);
		}
	}

	public void write(PrintWriter writer) {
		writer.print(makersEdges.getBitset());
		writer.print(',');
		writer.print(breakersEdges.getBitset());
		writer.println();
	}

	@Override
	public String toString() {
		return "Maker: " + makersEdges + "\n" + "Breaker: " + breakersEdges;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((breakersEdges == null) ? 0 : breakersEdges.hashCode());
		result = prime * result + ((freeEdges == null) ? 0 : freeEdges.hashCode());
		result = prime * result + ((makersEdges == null) ? 0 : makersEdges.hashCode());
		result = prime * result + ((usedEdges == null) ? 0 : usedEdges.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		CompleteGameGraph other = (CompleteGameGraph) obj;
		if (breakersEdges == null) {
			if (other.breakersEdges != null) {
				return false;
			}
		} else if (!breakersEdges.equals(other.breakersEdges)) {
			return false;
		}
		if (freeEdges == null) {
			if (other.freeEdges != null) {
				return false;
			}
		} else if (!freeEdges.equals(other.freeEdges)) {
			return false;
		}
		if (makersEdges == null) {
			if (other.makersEdges != null) {
				return false;
			}
		} else if (!makersEdges.equals(other.makersEdges)) {
			return false;
		}
		if (usedEdges == null) {
			if (other.usedEdges != null) {
				return false;
			}
		} else if (!usedEdges.equals(other.usedEdges)) {
			return false;
		}
		return true;
	}

}
