package rs.dmi.master.game.base;

import java.util.Comparator;

import rs.dmi.master.component.edge.Edge;
import rs.dmi.master.component.edge.EdgeCollection;
import rs.dmi.master.game.CompleteGameGraph;
import rs.dmi.master.game.GamePlay;
import rs.dmi.master.game.GameStatisticsLogger;
import rs.dmi.master.game.Player;
import rs.dmi.master.isomorphism.GameGraphState;
import rs.dmi.master.isomorphism.canonicallabel.AbstractCanonicalLabelingAlgorithm;
import rs.dmi.master.isomorphism.canonicallabel.NautyImpl;
import rs.dmi.master.lookuptable.LookupTable;
import rs.dmi.master.pathcollection.PathCollection;

/**
 * Base class for all games.
 * 
 * @author niktrk
 * @date 17.10.2012.
 * 
 * @param <T>
 *            - Type used for storage in path collection.
 */
public abstract class AbstractGame<T> {

	private CompleteGameGraph completeGameGraph;
	private LookupTable lookup;
	private int depth;
	private PathCollection<T> paths;
	private AbstractCanonicalLabelingAlgorithm canonicalLabelingAlgorithm;

	private Comparator<Edge> makerFreeEdgesComparator;
	private Comparator<Edge> breakerFreeEdgesComparator;

	private GameStatisticsLogger statistics;

	public AbstractGame(int size, GamePlay gamePlay, LookupTable lookup, PathCollection<T> pathCollection) {
		this(size, gamePlay, lookup, -1, pathCollection);
	}

	public AbstractGame(int size, GamePlay gamePlay, LookupTable lookup, int depth, PathCollection<T> pathCollection) {
		this(new CompleteGameGraph(size, gamePlay), lookup, depth, pathCollection);
	}

	public AbstractGame(CompleteGameGraph completeGameGraph, LookupTable lookup, int depth, PathCollection<T> pathCollection) {
		this.completeGameGraph = completeGameGraph;
		this.lookup = lookup;
		if (depth <= 0) {
			this.depth = completeGameGraph.getFreeEdges().getSize() + completeGameGraph.getUsedEdges().getSize();
		} else {
			this.depth = depth;
		}
		this.paths = pathCollection;
		this.statistics = new GameStatisticsLogger();
		// can be changed to DreadnautCaller for testing purpose
		this.canonicalLabelingAlgorithm = new NautyImpl();
		this.makerFreeEdgesComparator = getFreeEdgesComparatorForPlayer(Player.MAKER, completeGameGraph);
		this.breakerFreeEdgesComparator = getFreeEdgesComparatorForPlayer(Player.BREAKER, completeGameGraph);
		initPaths();
	}

	private void initPaths() {
		calculatePaths(completeGameGraph.getNumberOfVertices(), paths);
		statistics.onPathsCalculated(paths.size());
		for (Edge edge : completeGameGraph.getBreakersEdges()) {
			paths.removeAllPathsWithEdge(edge);
		}
	}

	public Player playTheGame() {
		statistics.onGameStart(getClass().getSimpleName(), completeGameGraph.getNumberOfVertices(), depth);
		Player winner = play(completeGameGraph, completeGameGraph.getStartPlayer());
		statistics.onGameEnd(winner);
		return winner;
	}

	private Player play(CompleteGameGraph completeGameGraph, Player currentPlayer) {
		statistics.onPlay();
		Player winner = null;

		GameGraphState graphState = null;
		if (completeGameGraph.getUsedEdges().getSize() <= depth) {
			graphState = canonicalLabelingAlgorithm.getCanonicalGraph(completeGameGraph);
			// graphState = new GameGraphState(completeGameGraph);
			winner = lookup.getWinner(graphState);
			if (winner != null) {
				return winner;
			}
		}

		Iterable<Edge> free;
		if (Player.BREAKER.equals(currentPlayer)) {
			free = completeGameGraph.getFreeEdges(breakerFreeEdgesComparator);
		} else {
			free = completeGameGraph.getFreeEdges(makerFreeEdgesComparator);
		}

		for (Edge edge : free) {

			// play it
			completeGameGraph.useEdge(edge, currentPlayer);

			T removed = null;
			if (Player.BREAKER.equals(currentPlayer)) {
				removed = paths.removeAllPathsWithEdge(edge);
			}

			// check winner
			if (isPlayerWinner(currentPlayer, edge)) {
				winner = currentPlayer;
			} else {
				winner = play(completeGameGraph, completeGameGraph.getNextPlayer());
			}

			// unplay it
			completeGameGraph.freeEdge(edge, currentPlayer);

			if (Player.BREAKER.equals(currentPlayer)) {
				paths.addPathCollection(removed);
			}

			// check halt
			if (winner.equals(currentPlayer)) {
				break;
			}
		}

		if (completeGameGraph.getUsedEdges().getSize() <= depth) {
			lookup.setWinner(graphState, winner);
			statistics.onNewLookupTableEntry();
		}

		return winner;
	}

	private boolean isPlayerWinner(Player player, Edge lastEdge) {
		if (Player.MAKER.equals(player)) {
			EdgeCollection m = completeGameGraph.getMakersEdges();
			return m.getSize() >= getMinimumNumberOfEdgesForMakerWin(completeGameGraph.getNumberOfVertices())
					&& paths.containsOrWillContainSubsetOf(m, completeGameGraph.getFreeEdges(), lastEdge);
		} else {
			return paths.isEmpty();
		}
	}

	/**
	 * By default return <code>null</code> which means that no sorting is going
	 * to be applied on free {@link Edge}s for both players.
	 */
	public Comparator<Edge> getFreeEdgesComparatorForPlayer(Player player, CompleteGameGraph gameGraph) {
		return null;
	}

	/**
	 * Should be numberOfVertices for "cycle game" and (numberOfVertices-1) for
	 * "path game".
	 */
	protected abstract int getMinimumNumberOfEdgesForMakerWin(int numberOfVertices);

	protected abstract void calculatePaths(int numberOfVertices, PathCollection<T> pathCollection);

}
