package rs.dmi.master.isomorphism;

import gnu.trove.map.TLongObjectMap;
import gnu.trove.map.hash.TLongObjectHashMap;
import gnu.trove.set.TLongSet;
import gnu.trove.set.hash.TLongHashSet;

import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import rs.dmi.master.component.edge.Edge;
import rs.dmi.master.game.CompleteGameGraph;
import rs.dmi.master.game.GamePlay;
import rs.dmi.master.game.Player;
import rs.dmi.master.isomorphism.canonicallabel.AbstractCanonicalLabelingAlgorithm;
import rs.dmi.master.isomorphism.canonicallabel.DreadnautCaller;
import rs.dmi.master.isomorphism.canonicallabel.NautyImpl;

/**
 * Used for generating all non isomorphic graphs.
 * 
 * @author niktrk
 * @date 23.12.2012.
 * 
 */
public class NonIsomorphicGraphGenerator {

	private CompleteGameGraph graph;
	private int depth;
	private boolean simpleGraphs;
	private PrintWriter printWriter;
	private TLongObjectMap<TLongSet>[] nonIsomorphicSet;
	private AbstractCanonicalLabelingAlgorithm canonicalLabelingAlgorithm;
	private int invocationCount;

	public NonIsomorphicGraphGenerator(int n, GamePlay gamePlay, int depth, boolean nativeNauty, boolean simpleGraphs, boolean store) {
		graph = new CompleteGameGraph(n, gamePlay);
		if (depth <= 0) {
			this.depth = graph.getFreeEdges().getSize();
		} else {
			this.depth = depth;
		}
		initCanonicalLabelingAlgorithm(nativeNauty);
		this.simpleGraphs = simpleGraphs;
		initPrintWriterIfNeeded(store);
		initNonIsomorphicSet();
		invocationCount = 0;
	}

	private void initCanonicalLabelingAlgorithm(boolean nativeNauty) {
		if (nativeNauty) {
			canonicalLabelingAlgorithm = new DreadnautCaller();
		} else {
			canonicalLabelingAlgorithm = new NautyImpl();
		}
	}

	private void initPrintWriterIfNeeded(boolean store) {
		if (store) {
			try {
				printWriter = new PrintWriter(new FileWriter(createFileName()));
			} catch (IOException e) {
				throw new IllegalStateException(e);
			}
		}
	}

	private String createFileName() {
		String name = "gen" + graph.getNumberOfVertices() + "-" + depth;
		if (simpleGraphs) {
			name += "simple-";
		} else {
			name += "normal-";
		}
		return name + canonicalLabelingAlgorithm.getName() + ".txt";
	}

	@SuppressWarnings("unchecked")
	private void initNonIsomorphicSet() {
		nonIsomorphicSet = new TLongObjectHashMap[depth + 1];
		for (int i = 0; i < nonIsomorphicSet.length; i++) {
			nonIsomorphicSet[i] = new TLongObjectHashMap<TLongSet>();
		}
	}

	public void startGeneration() {
		long start = System.currentTimeMillis();
		generate(graph, graph.getStartPlayer());
		System.out.println("Time: " + (System.currentTimeMillis() - start));
		if (printWriter != null) {
			printWriter.close();
		}
		printStatistics();
	}

	private void printStatistics() {
		int total = 0;
		for (int i = 0; i < nonIsomorphicSet.length; i++) {
			int subtotal = 0;
			for (TLongSet set : nonIsomorphicSet[i].valueCollection()) {
				subtotal += set.size();
			}
			System.out.println(i + ":  " + subtotal);
			total += subtotal;
		}
		System.out.println("Total number od non isomoprhic graphs: " + total);
		System.out.println("Canonical label algorithm invocations: " + invocationCount);
	}

	private void generate(CompleteGameGraph completeGameGraph, Player currentPlayer) {
		if (completeGameGraph.getFreeEdges().getSize() == 0 || completeGameGraph.getUsedEdges().getSize() == depth) {
			return;
		} else {
			for (Edge edge : completeGameGraph.getFreeEdges()) {
				completeGameGraph.useEdge(edge, currentPlayer);

				invocationCount++;
				GameGraphState graphState = canonicalLabelingAlgorithm.getCanonicalGraph(completeGameGraph);

				int i = completeGameGraph.getUsedEdges().getSize();
				long maker = graphState.getMakersEdges().getBitset();
				long breaker = graphState.getBreakersEdges().getBitset();
				TLongSet set = nonIsomorphicSet[i].get(maker);
				if (set == null) {
					set = new TLongHashSet();
					nonIsomorphicSet[i].put(maker, set);
				}

				if (!set.contains(breaker)) {
					set.add(breaker);
					if (printWriter != null) {
						completeGameGraph.write(printWriter);
					}
					if (simpleGraphs) {
						generate(completeGameGraph, currentPlayer);
					} else {
						generate(completeGameGraph, completeGameGraph.getNextPlayer());
					}
				}

				completeGameGraph.freeEdge(edge, currentPlayer);
			}
		}
	}

	public static void main(String[] args) {
		NonIsomorphicGraphGenerator gen = new NonIsomorphicGraphGenerator(7, GamePlay.UNBIASED_MAKER_STARTS, -1, false, false, false);
		gen.startGeneration();
	}

}
