package rs.dmi.master.isomorphism.canonicallabel;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import rs.dmi.master.component.edge.Edge;
import rs.dmi.master.game.CompleteGameGraph;
import rs.dmi.master.game.GamePlay;
import rs.dmi.master.game.Player;
import rs.dmi.master.isomorphism.GameGraphState;

/**
 * Java dreadnaut caller.
 * 
 * @author niktrk
 * @date 22.12.2012.
 * 
 */
public class DreadnautCaller extends AbstractCanonicalLabelingAlgorithm {

	private static final String SPACE = " ";
	private static final String SEMI_COLON = ";";

	// c - tells nauty to find a canonical labeling
	// -a - tells nauty not to display the automorphisms it finds
	// -m - tells nauty not to display the level markers
	// n=X - tells nauty that input graph has X vertices
	private static final String CANONICAL_LABEL_START = "c -a-m n=";

	// g - tells nauty to read the graph
	// example = g 7 4;8;9 3;10 2 4;11 0 3;12;13;0 9;1 10;2 7 11;3 8;4 9;5;6;
	// graph has edges: 0-7, 0-4, 1-8, 2-9, 2-3, 3-10, 3-2, 3-4, 4-11, ...
	private static final String CANONICAL_LABEL_MIDDLE = " g ";

	// x - execute nauty
	// b - display canonical label and the canonically labeled graph
	// canonical label is displayed as a list of the vertices in canonical order
	private static final String CANONICAL_LABEL_END = " x b";

	// f - tells nauty the initial partition
	// for n= 10, “f=[3:7 | 0,2]” is partition [3,4,5,6,7 | 0,2 | 1,8,9]
	private static final String CANONICAL_LABEL_INITIAL_PARTITION = "f=[ %d:%d | %d:%d ]";

	private ProcessBuilder dreadnautProcessBuilder;

	private static File DREADNAUT;
	static {
		try {
			DREADNAUT = new File(ClassLoader.getSystemResource("dreadnaut").toURI());
			Runtime.getRuntime().exec("chmod 777 " + DREADNAUT.getAbsolutePath()).waitFor();
		} catch (Exception e) {
			throw new IllegalStateException("Unable to prepare dreadnaut process.", e);
		}
	}

	public DreadnautCaller() {
		super();
		dreadnautProcessBuilder = new ProcessBuilder(DREADNAUT.getAbsolutePath());
		dreadnautProcessBuilder.redirectErrorStream(true);
	}

	private List<String> executeDreadnaut(String commands) {
		Process dreadnautProcess = null;
		try {
			dreadnautProcess = dreadnautProcessBuilder.start();

			BufferedWriter out = new BufferedWriter(new OutputStreamWriter(dreadnautProcess.getOutputStream()));
			out.write(commands);
			out.close();

			List<String> output = new ArrayList<String>();

			BufferedReader in = new BufferedReader(new InputStreamReader(dreadnautProcess.getInputStream()));
			String line = in.readLine();
			while (line != null) {
				output.add(line);
				line = in.readLine();
			}

			return output;
		} catch (Exception e) {
			throw new IllegalStateException(e);
		} finally {
			if (dreadnautProcess != null) {
				dreadnautProcess.destroy();
			}
		}
	}

	// adjacent vertices of i are in graph[i] in format x y z;
	private String[] convertToVertexColouredNautyGraph(CompleteGameGraph gameGraph) {
		// initialize
		int size = gameGraph.getNumberOfVertices();
		String[] graph = new String[size * 2];
		for (int i = 0; i < graph.length; i++) {
			// initialize all array elements to " "
			graph[i] = SPACE;
		}

		// add edges
		for (int i = 0; i < size; i++) {
			// add i -- i+size edge
			addEdge(graph, i, i + size);
		}
		for (Edge makersEdge : gameGraph.getMakersEdges()) {
			// add makers edges regularly
			addEdge(graph, makersEdge.getFirstVertex(), makersEdge.getSecondVertex());
		}
		for (Edge breakersEdge : gameGraph.getBreakersEdges()) {
			// add breakers edges shifted by size
			addEdge(graph, breakersEdge.getFirstVertex() + size, breakersEdge.getSecondVertex() + size);
		}

		// finalize
		for (int i = 0; i < graph.length; i++) {
			// add separators after each vertex
			graph[i] += SEMI_COLON;
		}

		return graph;
	}

	private void addEdge(String[] edges, int u, int v) {
		edges[u] = edges[u] + String.valueOf(v) + SPACE;
		edges[v] = edges[v] + String.valueOf(u) + SPACE;
	}

	private String generateCanonialLabelCommand(CompleteGameGraph completeGameGraph) {
		// generate nauty version of the graph
		String[] nautyGraph = convertToVertexColouredNautyGraph(completeGameGraph);

		// generate first part of the command and enter graph
		String command = CANONICAL_LABEL_START + nautyGraph.length + CANONICAL_LABEL_MIDDLE;
		for (int i = 0; i < nautyGraph.length; i++) {
			command += nautyGraph[i];
		}

		// generate initial partition command
		int size = completeGameGraph.getNumberOfVertices();
		command += String.format(CANONICAL_LABEL_INITIAL_PARTITION, 0, size - 1, size, 2 * size - 1);

		// generate execute and print label commands
		return command + CANONICAL_LABEL_END;
	}

	private int[] getCanonicalLabel(CompleteGameGraph completeGameGraph) {
		// generate dreadnaut command
		String command = generateCanonialLabelCommand(completeGameGraph);

		// execute dreadnaut and parse the response;
		// 4-th line of the output contains canonical label
		String[] mapping = executeDreadnaut(command).get(3).split(SPACE);

		// generate canonical label as int[]
		// there is a leading space so we start from index 1
		int[] canonicalLabel = new int[completeGameGraph.getNumberOfVertices()];
		for (int i = 1; i <= completeGameGraph.getNumberOfVertices(); i++) {
			canonicalLabel[Integer.valueOf(mapping[i])] = i - 1;
		}

		return canonicalLabel;
	}

	@Override
	public GameGraphState getCanonicalGraph(CompleteGameGraph completeGameGraph) {
		int[] map = getCanonicalLabel(completeGameGraph);
		return relabelGameGraph(completeGameGraph, map);
	}

	@Override
	public String getName() {
		return "NativeDreadnaut";
	}

	public static void main(String[] args) {
		CompleteGameGraph completeGameGraph = new CompleteGameGraph(7, GamePlay.UNBIASED_MAKER_STARTS);
		completeGameGraph.useEdge(Edge.EDGES[3][4], Player.MAKER);
		completeGameGraph.useEdge(Edge.EDGES[3][2], Player.MAKER);
		completeGameGraph.useEdge(Edge.EDGES[0][4], Player.MAKER);
		completeGameGraph.useEdge(Edge.EDGES[2][4], Player.BREAKER);
		completeGameGraph.useEdge(Edge.EDGES[0][2], Player.BREAKER);
		completeGameGraph.useEdge(Edge.EDGES[1][3], Player.BREAKER);

		DreadnautCaller caller = new DreadnautCaller();

		String command = caller.generateCanonialLabelCommand(completeGameGraph);
		System.out.println(command);
		List<String> executeDreadnaut = caller.executeDreadnaut(command);
		for (String string : executeDreadnaut) {
			System.out.println(string);
		}

		int[] canonicalLabel = caller.getCanonicalLabel(completeGameGraph);
		System.out.println(Arrays.toString(canonicalLabel));
		System.out.println(caller.getCanonicalGraph(completeGameGraph));
	}

}
